function [ei]=EdgeInfo(P,Node,Element,Domain)
%
%  Build data structures for cell-based discretizations on
%  polyhedral meshes.
%
%  INPUT:
%    P       : NElem x d array of cell centers
%    Node    : NNode x d array of mesh vertices
%    Element : Nelem x 1 cell structure, cell-to-vertex indices
%              (into Node)
%    Domain  : domain distance function (see comments in PolyMesher)
%
%  OUTPUT:
%    ei         : structure with the following fields
%    .EdgesI    : NEdgesI x 2 array of internal edge-to-vertex indices
%               in d>2 this should be a cell structure, similar to Element
%    .Edg2ElemI : NEdgesI x 2 array of internal edge-to-cell indices
%    .EdgLenI   : NEdgesI x 1 array of edge lengths (face areas)
%    .DistI     : NEdgesI x 2 array of cell-center-to-face distances
%    .SI        : NEdgesI x d array of face centroids
%    .ZI        : NEdgesI x d array of intersections between faces &
%                 lines connecting cell centers (these are
%                 different from face centroids)
%
%    .EdgesX    : same as .EdgesI, but for external edges
%    .Edg2ElemX : NEdgesX x 1 array of external edge-to-cell indices
%    .EdgLenX   : same as .EdgLenI, but for external edges
%    .DistX     : NEdgesX x 1 array of cell-center-to-face distances
%    .SX        : NEdgesX x d array of face centroids
%    .ZX        : NEdgesX x d array of intersections between faces &
%                 lines from centers, which are orthogonal to the face
%
NElem = size(Element,1);
% build a list of all edges & corresponding elements in the subdivision
alledges = cell2mat(...
    cellfun(@(elNodes) sort([elNodes([1:end])'; elNodes([2:end,1])'])', ...
            Element, 'UniformOutput',false));
e2el = cell2mat(...
    cellfun(@(elNodes,el) repmat(el,length(elNodes),1), ...
            Element, num2cell((1:NElem)'), 'UniformOutput',false));
% now figure out the unique edges
[Edg1,ix1,jx1]=unique(alledges,'rows','first');
[Edg2,ix2,jx2]=unique(alledges,'rows','last');
% Find external & internal edges
xe = find(ix1==ix2);
ie = find(ix1~=ix2);
NEdges=size(Edg1,1);
NEdgesI=length(ie);
NEdgesX=length(xe);
% Build data for the internal edges
ei.EdgesI    = Edg1(ie,:);
ei.Edg2ElemI = [e2el(ix1(ie)), e2el(ix2(ie))];
ei.EdgLenI   = sqrt( (Node(ei.EdgesI(:,1),1)-Node(ei.EdgesI(:,2),1)).^2 + ...
                     (Node(ei.EdgesI(:,1),2)-Node(ei.EdgesI(:,2),2)).^2   );
ei.SI        = (Node(ei.EdgesI(:,1),:)+Node(ei.EdgesI(:,2),:))/2;
ei.ZI        = line_intersect(Node(ei.EdgesI(:,1),:), ...
                              Node(ei.EdgesI(:,2),:), ...
                              P(ei.Edg2ElemI(:,1),:), ...
                              P(ei.Edg2ElemI(:,2),:));
ei.DistI   = [sqrt((P(ei.Edg2ElemI(1:NEdgesI,1),1)-ei.ZI(:,1)).^2+ ...
                   (P(ei.Edg2ElemI(1:NEdgesI,1),2)-ei.ZI(:,2)).^2), ...
              sqrt((P(ei.Edg2ElemI(1:NEdgesI,2),1)-ei.ZI(:,1)).^2+ ...
                   (P(ei.Edg2ElemI(1:NEdgesI,2),2)-ei.ZI(:,2)).^2)];
% Build data for the external edges
ei.EdgesX    = Edg1(xe,:);
ei.Edg2ElemX = e2el(ix1(xe));
ei.EdgLenX   = sqrt( (Node(ei.EdgesX(:,1),1)-Node(ei.EdgesX(:,2),1)).^2 + ...
                     (Node(ei.EdgesX(:,1),2)-Node(ei.EdgesX(:,2),2)).^2   );
ei.SX = (Node(ei.EdgesX(:,1),:)+Node(ei.EdgesX(:,2),:))/2;
dMX = Domain('Dist',ei.SX);
[d1,ix] = min(abs(dMX(:,1:end-1)),[],2);
d2  = Domain('Dist', P(ei.Edg2ElemX(:,1),:));
n1 = (Domain('Dist',ei.SX+repmat([eps,0],NEdgesX,1))-dMX)/eps;
n2 = (Domain('Dist',ei.SX+repmat([0,eps],NEdgesX,1))-dMX)/eps;
ei.ZX = P(ei.Edg2ElemX(:,1),:) - ...
    [n1(sub2ind(size(d2),(1:NEdgesX)',ix)).*d2(sub2ind(size(d2),(1:NEdgesX)',ix)), ...
     n2(sub2ind(size(d2),(1:NEdgesX)',ix)).*d2(sub2ind(size(d2),(1:NEdgesX)',ix))];
ei.DistX  = abs(d2(sub2ind(size(d2),(1:NEdgesX)',ix)));
%--------------------------------------------------------------------------
function [xy]=line_intersect(xy1,xy2,xy3,xy4)
% compute coordinates of the point where two line segments 
% (xy1, xy2) and (xy3, xy4) intersect
x1 = xy1(:,1); y1 = xy1(:,2);
x2 = xy2(:,1); y2 = xy2(:,2);
x3 = xy3(:,1); y3 = xy3(:,2);
x4 = xy4(:,1); y4 = xy4(:,2);
%
det=(x1-x2).*(y3-y4) - (y1-y2).*(x3-x4);
xy=[((x1.*y2-y1.*x2).*(x3-x4)-(x1-x2).*(x3.*y4-y3.*x4))./det, ...
    ((x1.*y2-y1.*x2).*(y3-y4)-(y1-y2).*(x3.*y4-y3.*x4))./det];
