close all;
%
% Form a polyhedral cell based FVM discretization of Laplace PDE
%
%
% 1. Provide the domain description, run the mesher, and compute
% some additional mesh-related structures
%
Ncells    = 400;
mesh_iter = 40;
%
Domain = @UnitSquareDomain;
%Domain = @UnitCircleDomain;
%Domain = @WrenchDomain;
[Node,Element,~,~,P] = PolyMesher(Domain, Ncells, mesh_iter);
[ei] = EdgeInfo(P,Node,Element,Domain);
% Compute cell areas - is also used for integration
EA = cellfun(@(Ev) ...
             0.5*( ...
                 Node(Ev,1)'*circshift(Node(Ev,2),-1) ...
                 - ...
                 Node(Ev,2)'*circshift(Node(Ev,1),-1) ...
                 ), Element, 'UniformOutput', true);
%
% 2. This is used both for forming the system and testing the code.
% Basically we are solving the Dirichlet boundary value problem
%
%    -[y_xx + y_yy] = f_exact  in Omega
%      y = y_exact             on Gamma
%
y_exact = @(xy) cos(4.0*pi*xy(:,1)) + sin(2.0*pi*xy(:,2));
f_exact = @(xy) 16.0*pi*pi*cos(4.0*pi*xy(:,1)) + 4.0*pi*pi*sin(2.0*pi*xy(:,2));
%
% We then evaluate the right hand side at cell centers, and the
% boundary condition at face centers on the boundary.
% In a control problem F and G may of course also be discrete unknowns.
%
F = f_exact(P);
G = y_exact(ei.ZX);
%
% 3. Form the system matrix; A = discrete representation of the
% Laplace operator on our grid
%
ndof = size(Element,1);  % number of unknowns = number of cells
%
% integrate over internal edges
Ii=reshape(...
    [ei.Edg2ElemI(:,1),ei.Edg2ElemI(:,1),   ...
     ei.Edg2ElemI(:,2),ei.Edg2ElemI(:,2)]', ...
    4*size(ei.EdgesI,1),1);
Ji=reshape(...
    [ei.Edg2ElemI(:,1),ei.Edg2ElemI(:,2),   ...
     ei.Edg2ElemI(:,1),ei.Edg2ElemI(:,2)]', ...
    4*size(ei.EdgesI,1),1);
tau = ei.EdgLenI./sum(ei.DistI,2);
Si=reshape(...
    [tau,-tau,-tau,tau]', ...
    4*size(ei.EdgesI,1),1);
% integrate over external edges
Ix=ei.Edg2ElemX;
Jx=ei.Edg2ElemX;
Sx=ei.EdgLenX./ei.DistX;
% assemble everything into the matrix
A = sparse([Ii;Ix], [Ji;Jx], [Si;Sx], ndof, ndof);
%
% 4. Form the right hand side
%
b = F .* EA + ...
    accumarray(ei.Edg2ElemX,G.*ei.EdgLenX./ei.DistX,[ndof,1]);
%
% 5. Solve the system
%
Y = A\b;
%
% 6. Vizualize the solution
%
clf; axis equal; axis off; hold on;
MaxNVer = max(cellfun(@numel,Element));      %Max. num. of vertices in mesh
PadWNaN = @(E) [E(:)' NaN(1,MaxNVer-numel(E))];  %Pad cells with NaN
ElemMat = cellfun(PadWNaN,Element,'UniformOutput',false);
ElemMat = vertcat(ElemMat{:});               %Create padded element matrix
patch('Faces',ElemMat,'Vertices',Node,'FaceColor','flat','FaceVertexCData',Y);
colorbar;
%
% 7. For testing purposes, we also evaluate the exact solution
% (which we know for this artificial problem) at the cell centers
% and compute the L2-error norm.
%
Y_exact = y_exact(P);
L2_err  = sqrt(sum((Y-Y_exact).^2.*EA));
% Calculate the largest cell size (our grid is quasi-uniform here
% any ways), assume h ~ area^0.5 in 2D.
h = sqrt(max(EA));
disp(sprintf('h = %15.6e, L2 error = %15.6e',h,L2_err));
